/*****************************************************************************
 * colorspace.c: misc colorspace conversion functions
 *****************************************************************************
 * $Id: colorspace.c,v 1.6 2004/04/09 08:14:34 alainjj Exp $
 *****************************************************************************
 * Copyright (C) 1998 Gerd Knorr <kraxel@cs.tu-berlin.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 *****************************************************************************
 *
 * misc colorspace conversion functions
 * most of them have common arguments (wanna be able to use function
 * pointers):
 *     foo(unsigned char* dest, unsigned char* src, int width, int height);
 *
 *****************************************************************************/
#include "config.h"
#include "colorspace.h"

#define CLIP         320

#if 0
# define RED_NULL    137
# define BLUE_NULL   156
# define LUN_MUL     360
# define RED_MUL     512
# define BLUE_MUL    512
#else
# define RED_NULL    128
# define BLUE_NULL   128
# define LUN_MUL     256
# define RED_MUL     512
# define BLUE_MUL    512
#endif

#define GREEN1_MUL  (-RED_MUL/2)
#define GREEN2_MUL  (-BLUE_MUL/6)
#define RED_ADD     (-RED_NULL  * RED_MUL)
#define BLUE_ADD    (-BLUE_NULL * BLUE_MUL)
#define GREEN1_ADD  (-RED_ADD/2)
#define GREEN2_ADD  (-BLUE_ADD/6)

/* lookup tables */
static unsigned int  ng_yuv_gray[256];
static unsigned int  ng_yuv_red[256];
static unsigned int  ng_yuv_blue[256];
static unsigned int  ng_yuv_g1[256];
static unsigned int  ng_yuv_g2[256];
static unsigned int  ng_clip[256 + 2 * CLIP];

#define GRAY(val)            ng_yuv_gray[val]
#define RED(gray,red)        ng_clip[ CLIP + gray + ng_yuv_red[red] ]
#define GREEN(gray,red,blue) ng_clip[ CLIP + gray + ng_yuv_g1[red] +	\
						              ng_yuv_g2[blue] ]
#define BLUE(gray,blue)      ng_clip[ CLIP + gray + ng_yuv_blue[blue] ]

/* ------------------------------------------------------------------- */
/* RGB conversions                                                     */

void
swap_rgb16 (char *mem, int n)
{
  char c;
  char *p = mem;
  int i = n;

  while (i--)
    {
      c = p[0];
      p[0] = p[1];
      p[1] = c;
      p += 2;
    }
}

void
swap_rgb24 (char *mem, int n)
{
  char c;
  char *p = mem;
  int i = n;

  while (i--)
    {
      c = p[0];
      p[0] = p[2];
      p[2] = c;
      p += 3;
    }
}

void
swap_rgb32 (char *mem, int n)
{
  char c;
  char *p = mem;
  int i = n;

  while (i--)
    {
      c = p[0];
      p[0] = p[3];
      p[3] = c;
      c = p[1];
      p[1] = p[2];
      p[2] = c;
      p += 4;
    }
}

void
rgb24_to_bgr24 (char *dest, char *src, int w, int h)
{
  register char *s = src;
  register char *d = dest;
  register int i = w * h;

  while (i--)
    {
      *(d++) = s[2];
      *(d++) = s[1];
      *(d++) = s[0];
      s += 3;
    }
}

void
rgb24_to_rgb32 (unsigned char *dest, unsigned char *src, int w, int h)
{
  int i = w * h;

  while (i--)
    {
#ifdef WORDS_BIGENDIAN
      *(dest++) = 0;
#endif
      *(dest++) = *(src++);
      *(dest++) = *(src++);
      *(dest++) = *(src++);
#ifndef WORDS_BIGENDIAN
      *(dest++) = 0;
#endif
    }
}

void
rgb32_to_rgb24 (unsigned char *dest, unsigned char *src, int w, int h)
{
  int i = w * h;

  while (i--)
    {
#ifdef WORDS_BIGENDIAN
      src++;
#endif
      *(dest++) = *(src++);
      *(dest++) = *(src++);
      *(dest++) = *(src++);
#ifndef WORDS_BIGENDIAN
      src++;
#endif
    }
}

/* ------------------------------------------------------------------- */
/* YUV conversions                                                     */


void
packed422_to_planar422 (unsigned char *d, unsigned char *s, int w, int h)
{
  int i;
  unsigned char *y, *u, *v;

  i = (w * h) / 2;
  y = d;
  u = y + w * h;
  v = u + w * h / 2;

  while (i--)
    {
      *(y++) = *(s++);
      *(u++) = *(s++);
      *(y++) = *(s++);
      *(v++) = *(s++);
    }
}

void
packed422_to_planar411 (unsigned char *d, unsigned char *s, int w, int h)
{
  int a, b;
  unsigned char *y, *u, *v;

  y = d;
  u = y + w * h;
  v = u + w * h / 4;

  for (a = h; a > 0; a -= 2)
    {
      for (b = w; b > 0; b -= 2)
        {

          *(y++) = *(s++);
          *(u++) = *(s++);
          *(y++) = *(s++);
          *(v++) = *(s++);
        }
      for (b = w; b > 0; b -= 2)
        {
          *(y++) = *(s++);
          s++;
          *(y++) = *(s++);
          s++;
        }
    }
}

/* ------------------------------------------------------------------- */
/* YUV to RGB conversions                                              */

void
yuv422_to_rgb24(unsigned char *dest, unsigned char *s, int p)
{
    unsigned char *d = dest;
    int gray;
    
    while (p) {
	gray = GRAY(s[0]);
	d[0] = RED(gray,s[3]);
	d[1] = GREEN(gray,s[3],s[1]);
	d[2] = BLUE(gray,s[1]);
	gray = GRAY(s[2]);
	d[3] = RED(gray,s[3]);
	d[4] = GREEN(gray,s[3],s[1]);
	d[5] = BLUE(gray,s[1]);
	d += 6;
	s += 4;
	p -= 2;
    }
}

void yuv2rgb_init(void)
{
    int i;
    
    /* init Lookup tables */
    for (i = 0; i < 256; i++) {
        ng_yuv_gray[i] = i * LUN_MUL >> 8;
        ng_yuv_red[i]  = (RED_ADD    + i * RED_MUL)    >> 8;
        ng_yuv_blue[i] = (BLUE_ADD   + i * BLUE_MUL)   >> 8;
        ng_yuv_g1[i]   = (GREEN1_ADD + i * GREEN1_MUL) >> 8;
        ng_yuv_g2[i]   = (GREEN2_ADD + i * GREEN2_MUL) >> 8;
    }
    for (i = 0; i < CLIP; i++)
        ng_clip[i] = 0;
    for (; i < CLIP + 256; i++)
        ng_clip[i] = i - CLIP;
    for (; i < 2 * CLIP + 256; i++)
        ng_clip[i] = 255;

}
